/**
 * copyright area
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.metadata.SpeedoXMLDescriptor;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Attribute;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Collection;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;

public class AroundCompilation extends EnhancerComponent {

    public final static String LOGGER_NAME
        = SpeedoProperties.LOGGER_NAME + ".generation.enhancer.precompile";

    boolean beforeCompilation;

    public AroundCompilation(boolean beforeCompilation) {
        this.beforeCompilation = beforeCompilation;
    }

    public boolean init() throws SpeedoException {
        logger = scp.loggerFactory.getLogger(LOGGER_NAME);
		debug = logger.isLoggable(BasicLevel.DEBUG);
	    return !scp.getXmldescriptor().isEmpty();
    }

    public void process() throws SpeedoException {
        if (scp.getXmldescriptor() == null || scp.getXmldescriptor().isEmpty()) {
            return;
        }
        Collection xmls = scp.getXmldescriptor().values();
        for (Iterator itDesc = xmls.iterator(); itDesc.hasNext();) {
            SpeedoXMLDescriptor desc = (SpeedoXMLDescriptor) itDesc.next();
            for (Iterator itclass = desc.getSpeedoClasses().iterator(); itclass.hasNext();) {
                SpeedoClass jdoclass = (SpeedoClass) itclass.next();
				if (debug) {
					logger.log(BasicLevel.DEBUG,
							"Arround the compilation of the class "
							+ jdoclass.getFQName() + ": "
							+ (beforeCompilation ? "before" : "after")
							+ " the compilation");
				}
                ClassWriter cw = new ClassWriter(false);
                InheritanceModifier im = new InheritanceModifier(
                    cw, logger, jdoclass, beforeCompilation);
				DummyAccessorModifier dam = new DummyAccessorModifier(
					im, logger, jdoclass, beforeCompilation);
                String name = jdoclass.getFQName();
                ClassReader cr = loadJavaClass(false, name, scp.output, false);
                if (beforeCompilation) {
					EmptyConstructorAdder eca = new EmptyConstructorAdder(
						dam, logger, jdoclass.noArgConstructorStatus);
					cr.accept(eca, false);
				} else {
					cr.accept(dam, false);
				}
                writeJavaClass(name, cw, scp.output);
            }
        }
    }
}
