package org.objectweb.speedo.generation.enhancer;

import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.metadata.SpeedoXMLDescriptor;
import org.objectweb.speedo.metadata.SpeedoPackage;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.generation.mivisitor.MetaInfoVisitor;
import org.objectweb.speedo.generation.mivisitor.AbstractMetaInfoVisitor;
import org.objectweb.speedo.generation.api.SpeedoCompilerParameter;
import org.objectweb.asm.ClassReader;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

public class MetaDataAnalyzer extends  AbstractMetaInfoVisitor {

    public final static String LOGGER_NAME
            = SpeedoProperties.LOGGER_NAME + ".generation.enhancer.analyzer";

    private Loader loader;

    public MetaDataAnalyzer() {
        super();
        loader = new Loader();
    }

    public MetaDataAnalyzer(MetaInfoVisitor mim) {
        super(mim);
        loader = new Loader();
    }

    private class Loader extends EnhancerComponent {

        public boolean init() throws SpeedoException {
            isSrcJar = false;
	        return true;
        }

        public void process() throws SpeedoException {
            //nothing
        }
        public void setLogger(Logger logger) {
            this.logger = logger;
        }
    }

    protected String getLoggerName() {
        return LOGGER_NAME;
    }

    public void setLogger(Logger logger) {
        super.setLogger(logger);
        loader.setLogger(logger);
    }

    public void visitCompilerParameter(SpeedoCompilerParameter scp) throws SpeedoException {
        loader.setSpeedoCompilerParameter(scp);
        super.visitCompilerParameter(scp);
    }

    public void visitClass(SpeedoClass sc) throws SpeedoException {
        logger.log(BasicLevel.DEBUG, "MetaDataAnalyzer.visitClass(" + sc.getFQName() + ")");
        boolean isSrcJar = false;
        SpeedoCompilerParameter scp = loader.getSpeedoCompilerParameter();
	    ClassReader cr = loader.loadJavaClass(
                    isSrcJar,
                    sc.jdoPackage.name + '.' + sc.name,
                    scp.output,
                    false);

        cr.accept(new ClassInfoAnalyzer(
                loader, sc, scp.getXmldescriptor().values(), logger), true);
        super.visitClass(sc);
    }
}
