/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.mapper.lib;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PClassMappingCtrl;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.lib.BasicClass;
import org.objectweb.jorm.metainfo.lib.BasicCompositeName;
import org.objectweb.jorm.metainfo.lib.JormManager;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.lib.KFPNCManager;
import org.objectweb.jorm.lib.JormPathHelper;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.speedo.mapper.api.JormFactory;
import org.objectweb.speedo.genclass.api.SpeedoGenClassProxy;
import org.objectweb.speedo.mim.api.SpeedoProxy;
import org.objectweb.speedo.mapper.api.JormFactoryAttributes;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.speedo.naming.lib.NamingManagerFactory;
import org.objectweb.speedo.naming.api.NamingManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.RegisterClassListener;
import javax.jdo.spi.JDOImplHelper;
import javax.jdo.spi.RegisterClassEvent;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Date;
import java.util.Properties;
import java.math.BigInteger;
import java.math.BigDecimal;

/**
 * This class manages the initialization and the naming, of persistent class.
 * The type of naming supported depends on the NamingManager instance managed
 * by the NamingManager
 *
 * @author S.Chassande-Barrioz
 */
public class BasicJormFactory
		implements JormFactory,
	PNameCoder,
	BindingController,
	LifeCycleController,
	JormFactoryAttributes,
	RegisterClassListener {

	/**
	 * fractal binding name to the mapper
	 */
	public final static String MAPPER_BINDING = "mapper";
	/**
	 * fractal binding name to the cache manager
	 */
	public final static String CACHE_MANAGER_BINDING = "cache-manager";

	/**
	 * The mapper is used to map and find the Jorm classes.
	 */
	protected PMapper mapper = null;

	/**
	 * The cache to assign to the binders
	 */
	protected CacheManager cache = null;

	/**
	 * key = String class name
	 * value = PBinder binder of the class
	 */
	protected Map binders = null;

	/**
	 * key = String class name
	 * value = PNamingContext pnc of the class
	 */
	protected Map pnamingContexts = null;
	/**
	 * temporaly variable use in the recursive algorithm
	 */
	protected Map cn2pcm = null;

	/**
	 * The manager of NamingManager available.
	 */
	private NamingManagerFactory nmf;

	/**
	 * The rule concerning the data structure (jorm property of the mapper)
	 */
	private byte mappingStructuresRule = JormFactoryAttributes.CREATE_IF_REQUIRED;

	protected Logger logger = null;

	private KFPNCManager kfpncManager;

	/**
	 * builds the BasicJormFactory
	 * and the hosting structure for the binder list.
	 */
	public BasicJormFactory() {
		pnamingContexts = new HashMap();
		binders = new HashMap();
		cn2pcm = new HashMap();
		nmf = new NamingManagerFactory();
		kfpncManager = new KFPNCManager(pnamingContexts);
	}

	public PBinder getPBinder(String className) {
		return (PBinder) binders.get(className);
	}

	// IMPLEMENTATION OF THE JormFactoryAttributes INTERFACE //
	//-------------------------------------------------------//

	public byte getMappingStructureRule() {
		return mappingStructuresRule;
	}

	public void setMappingStructureRule(byte rule) {
		mappingStructuresRule = rule;
	}

	// IMPLEMENTATION OF THE UserBindingController INTERFACE //
	//-------------------------------------------------------//

	public String[] listFc() {
		return new String[]{
			MAPPER_BINDING,
			CACHE_MANAGER_BINDING
		};
	}

	public Object lookupFc(String s) {
		if (MAPPER_BINDING.equals(s))
			return mapper;
		else if (CACHE_MANAGER_BINDING.equals(s))
			return cache;
		else
			return null;
	}

	public void bindFc(String s, Object o) {
		if ("logger".equals(s)) {
			logger = (Logger) o;
		} else if ("monolog-factory".equals(s)) {
			nmf.setLogger(((LoggerFactory) o).getLogger(logger.getName() + ".naming-mmanager-factory"));
			kfpncManager.setLogger(((LoggerFactory) o).getLogger(logger.getName() + ".kfpncmanager"));
		} else if (MAPPER_BINDING.equals(s)) {
			mapper = (PMapper) o;
		} else if (CACHE_MANAGER_BINDING.equals(s)) {
			cache = (CacheManager) o;
		}
	}

	public void unbindFc(String s) {
		if (MAPPER_BINDING.equals(s)) {
			mapper = null;
			nmf.setMapper(null);
		} else if (CACHE_MANAGER_BINDING.equals(s)) {
			cache = null;
		}
	}

	private boolean started = false;

	public String getFcState() {
		return started ? STARTED : STOPPED;
	}

	public void startFc() throws IllegalLifeCycleException {
		if (!started) {
			mapper.addMapperEventListener(kfpncManager);
			nmf.setMapper(mapper);
			JDOImplHelper.getInstance().addRegisterClassListener(this);
			started = true;
		}
	}

	public void stopFc() throws IllegalLifeCycleException {
		if (started) {
			mapper.removeMapperEventListener(kfpncManager);
			JDOImplHelper.getInstance().removeRegisterClassListener(this);
			started = false;
		}
	}

	// IMPLEMENTATION OF THE JormFactory INTERFACE //
	//---------------------------------------------//

	/**
	 * @return the PBinder instance which manages the identifer of the
	 * SpeedoProxy instance
	 */
	public PBinder getPBinder(SpeedoProxy sp) throws PException {
		String clName = sp.getClass().getName();
		PClassMapping pcm = mapper.lookup(clName);
		if (pcm != null)
			return pcm.getPBinder();
		return getPClassMapping(sp).getPBinder();
	}

	/**
	 * This Method is a shortcut to the getPBinder(SpeedoProxy) method. It only
	 * does the instanciation of the classname
	 * (Class.ForName(classsName).newInstance()) and call the
	 * getPBinder(SpeedoProxy) method with the created instance. Then if an
	 * instance is availlable it is better to use the other method.
	 * @param clazz the Jorm class name managed by the wanted PBinder
	 * @return the PBinder instance to use for the given jorm class name
	 * @throws org.objectweb.jorm.api.PException
	 */
	public PBinder getPBinder(Class clazz) throws PException {
		String clName = clazz.getName();
		PClassMapping pcm = mapper.lookup(clName);
		if (pcm != null)
			return pcm.getPBinder();
		return getPClassMapping(clazz).getPBinder();
	}

	public PBinder getPBinder(String classname, ClassLoader cl) throws PException {
		PClassMapping pcm = mapper.lookup(classname);
		if (pcm != null)
			return pcm.getPBinder();
		return getPClassMapping(getClass(classname, cl)).getPBinder();
	}

	public ClassLoader getClassLoader(String className) {
		Object o = mapper.lookup(className);
		if (o == null) {
			return null;
		}
		return getClassLoader(o.getClass());
	}

	/**
	 * Retrieves a PClassMapping instance managing a SpeedoProxy. If the
	 * PClassMapping does not already exist then it is created, configured and
	 * mapped in the PMapper. The PClassMapping configuration contains the
	 * PBinder assignement, the PNamingContext assignement for each reference,
	 * and the GenClassMapping assignement. The assigned GenClassMapping must
	 * also be configured. In other hand the PClassMapping of referenced class
	 * is not configured now. Only the PNamingContext is needed.
	 * @param sp is the SpeedoProxy instance which the PClassMapping is required
	 * @return the PClassMapping which manages a SpeedoProxy
	 * @throws org.objectweb.jorm.api.PException
	 */
	public PClassMapping getPClassMapping(SpeedoProxy sp) throws PException {
		if (sp instanceof SpeedoGenClassProxy) {
			// The GenClassMapping must be already created before
			SpeedoGenClassProxy sgcp = (SpeedoGenClassProxy) sp;
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
						"Looking for the PClassMapping of the generic class "
						+ sgcp.jdoGetGenClassId());
			}
			return JormPathHelper.getPClassMapping(sgcp.jdoGetGenClassId(), mapper);
		}
		Class spClass = sp.getClass();
		return getPClassMapping(spClass.getName(),
			spClass.getClassLoader(),
			getClassProperties(spClass));
	}

	private Properties getClassProperties(Class spClass) throws PException {
		try {
			return (Properties) spClass
				.getMethod("jdoGetClassProperties", new Class[]{})
				.invoke(null, new Object[]{});
		} catch (Exception e) {
			throw new PException(e,
				"Impossible to find the naming properties of the class "
				+ spClass.getName());
		}
	}

	public PClassMapping getPClassMapping(String className,
										  ClassLoader classLoader,
										  Properties classProperties) throws PException {
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		PClassMapping pcm = mapper.lookup(className);
		if (pcm == null) {
			synchronized (this) {
				pcm = (PClassMapping) cn2pcm.get(className);
				if (pcm != null) {
					return pcm;
				}
				pcm = mapper.lookup(className);
				if (pcm == null) {
					// Creates the PClassMapping, its PBinder and the PNC
					if (debug) {
						logger.log(BasicLevel.DEBUG,
								"Looking for the PClassMapping of the class " + className);
					}
					String pcmcn = NamingRules.fqMappingName(className);
					try {
						pcm = (PClassMapping) classLoader.loadClass(pcmcn).newInstance();
					} catch (Exception e) {
						throw new PException(e,
								"Impossible to instanciate the PClassMapping of the class "
								+ className + ": " + pcmcn);
					}
					if (debug) {
						logger.log(BasicLevel.DEBUG, "Jorm config:" + classProperties);
					}

					loadJormMI(className, classProperties, classLoader);
					// Fetch the Binder by the DataStoreNaming
					PBinder binder = findPBinder(
							classProperties.getProperty(JormPathHelper.getPath(className)),
							className, classLoader);
					pcm.setPBinder(binder);

					PNameCoder classPnc = findPNameManager(
								classProperties.getProperty(JormPathHelper.getPath(className)),
								className, classLoader);
					((PClassMappingCtrl) pcm).setClassPNameCoder(classPnc);

					binder.setPClassMapping(pcm);
					//Assign the PNamingContext of the reference
					RefConfig refConfig = new RefConfig(classProperties, mapper.getMapperName(), classLoader);
					cn2pcm.put(className, pcm);
					pcm.configureRefFields(refConfig);
					cn2pcm.remove(className);
					if (debug) {
						logger.log(BasicLevel.DEBUG, "PClassMapping/PBinder/PNC created for the class " + className);
						logger.log(BasicLevel.DEBUG, "- pcm: " + pcm);
						logger.log(BasicLevel.DEBUG, "- binder: " + binder);
					}

					// do the 'Map' operation on the mapper and register the pcm
					// in the local map
					mapper.map(null, pcm);
					logger.log(BasicLevel.INFO, "Class " + className + " mapped");
					PMapCluster cluster = mapper.getPMappingStructuresManager().getPMapCluster(className);
					if (cluster.isDefined()) {
						if (mappingStructuresRule >= CREATE_IF_REQUIRED) {
							cluster.createMappingStructures(
									mappingStructuresRule == FORCE_CREATE);
							if (mappingStructuresRule == DELETE_DATA) {
								cluster.deleteData();
							}
						}
						if (logger.isLoggable(BasicLevel.INFO)) {
							logger.log(BasicLevel.INFO, "Cluster of Class "
									+ cluster.getClusterClasses()
									+ " initialized ("
									+ mappingStructuresRule + ").");
						}
					} else {
						getPClassMapping((String) cluster
								.getUnResolvedDependencies().iterator().next(),
								classLoader);
					}
				} else if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "PClassMapping for class " + className
							+ " already exist:" + pcm);
			}
		}
		return pcm;
	}

	/**
	 * This Method is a shortcut to the getPClassMapping(SpeedoProxy) method. It only
	 * does the instanciation of the classname
	 * (Class.ForName(classsName).newInstance()) and call the
	 * getPClassMapping(SpeedoProxy) method with the created instance. Then if an
	 * instance is availlable it is better to use the other method.
	 *
	 * Retrieves a PClassMapping instance managing a jorm class. If the
	 * PClassMapping does not already exist then it is created, configured and
	 * mapped in the PMapper. The PClassMapping configuration contains the
	 * PBinder assignement, the PNamingContext assignement for each reference,
	 * and the GenClassMapping assignement. The assigned GenClassMapping must
	 * also be configured. In other hand the PClassMapping of referenced class
	 * is not configured now. Only the PNamingContext is needed.
	 * @param classname is the class name of the jorm class which the
	 * PClassMapping is required
	 * @return the PClassMapping which manages a SpeedoProxy
	 * @throws org.objectweb.jorm.api.PException
	 */
	public PClassMapping getPClassMapping(String classname, ClassLoader cl) throws PException {
		PClassMapping pcm = mapper.lookup(classname);
		if (pcm == null)
			pcm = getPClassMapping(getClass(classname, cl));
		return pcm;
	}

	public PClassMapping getPClassMapping(Class clazz) throws PException {
		PClassMapping pcm = mapper.lookup(clazz.getName());
		if (pcm == null) {
			pcm = getPClassMapping(clazz.getName(),
				clazz.getClassLoader(),
				getClassProperties(clazz));
		}
		return pcm;
	}

	/**
	 * @return the PNamingContext instance which manages a reference to the
	 * SpeedoProxy instance
	 */
	public PNamingContext getPNamingContext(SpeedoProxy sp) throws PException {
		return getPNamingContext(sp.getClass());
	}

	/**
	 * This Method is a shortcut to the getPNamingContext(SpeedoProxy) method.
	 * It only does the instanciation of the classname
	 * (Class.ForName(classsName).newInstance()) and call the
	 * getPNamingContext(SpeedoProxy) method with the created instance. Then if
	 * an instance is availlable it is better to use the other method.
	 * @param classname the Jorm class name managed by the wanted PNamingContext
	 * @return the PNamingContext instance to use for the given jorm class name
	 * @throws org.objectweb.jorm.api.PException
	 */
	public PNamingContext getPNamingContext(String classname, ClassLoader cl) throws PException {
		return getPNamingContext(getClass(classname, cl));
	}

	public PNamingContext getPNamingContext(Class clazz) throws PException {
		String className = clazz.getName();
		String path = JormPathHelper.getPath(className);
		Properties classProperties = getClassProperties(clazz);
		String  jormConf = classProperties.getProperty(path);
		if (jormConf != null) {
			return (PNamingContext) findPNameManager(
					jormConf, className, clazz.getClassLoader());
		} else {
			throw new PException(
					"Impossible to find the PNamingContext for the class "
					+ className);
		}
	}

	// IMPLEMENTATION OF THE RegisterClassListener INTERFACE //
	//-------------------------------------------------------//

	public void registerClass(RegisterClassEvent event) {
		try {
			getPClassMapping(event.getRegisteredClass());
		} catch (PException e) {
			throw new JDOException(
				"Impossible to initialize the persistent class '"
				+ event.getRegisteredClass().getName() + "':", e);
		}
	}


	// IMPLEMENTATION OF THE PNameCoder INTERFACE //
	//--------------------------------------------//
	public boolean codingSupported(int codingtype) {
		return false;
	}

	public PName decode(byte[] en) throws PExceptionNaming {
		return null;
	}

	public PName decodeAbstract(Object oid, Object context)
			throws PExceptionNaming, UnsupportedOperationException {
		Class clazz = (Class) context;
		if (oid instanceof PName) {
			try {
				return ((PName) oid).resolve(null);
			} catch (PException e) {
				throw new JDOUserException("Impossible to decode the identifier "
						+ oid + (clazz == null
						? "" : " of the class " + clazz.getName()), e);
			}
		}
		PBinder binder = null;
		PName pn = null;
		try {
			if (clazz != null) {
				binder = getPBinder(clazz);
			}
			pn = nmf.decode(binder, oid, clazz, this);
		} catch (PException e) {
			throw new JDOUserException("Impossible to decode the identifier "
					+ oid + (clazz == null
					? "" : " of the class " + clazz.getName()), e);
		}
		if (pn != null) {
			return pn;
		}
		if (binder != null) {
			return binder.decodeAbstract(oid, null);
		}
		throw new JDOUserException("Impossible to decode the identifier "
				+ oid + (clazz == null
				? "" : " of the class " + clazz.getName()));
	}

	public PName decodeByte(byte en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeObyte(Byte en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeChar(char en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeOchar(Character en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeInt(int en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeOint(Integer en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeLong(long en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeOlong(Long en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeShort(short en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeOshort(Short en) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public PName decodeString(String en) throws PExceptionNaming {
		return null;
	}

	public PName decodeCharArray(char[] en) throws PExceptionNaming {
		return null;
	}

	public PName decodeDate(Date en) throws PExceptionNaming {
		return null;
	}

	public PName decodeBigInteger(BigInteger en) throws PExceptionNaming {
		return null;
	}

	public PName decodeBigDecimal(BigDecimal en) throws PExceptionNaming {
		return null;
	}

	public byte[] encode(PName pn) throws PExceptionNaming {
		return new byte[0];
	}

	public Object encodeAbstract(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		try {
			return nmf.encode(pn);
		} catch (PException e) {
			throw new JDOUserException(
					"Impossible to encode the identifier :" + pn);
		}
	}

	public byte encodeByte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return 0;
	}

	public Byte encodeObyte(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public char encodeChar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return 0;
	}

	public Character encodeOchar(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public int encodeInt(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return 0;
	}

	public Integer encodeOint(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public long encodeLong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return 0;
	}

	public Long encodeOlong(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public short encodeShort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return 0;
	}

	public Short encodeOshort(PName pn) throws PExceptionNaming, UnsupportedOperationException {
		return null;
	}

	public String encodeString(PName pn) throws PExceptionNaming {
		return null;
	}

	public char[] encodeCharArray(PName pn) throws PExceptionNaming {
		return new char[0];
	}

	public Date encodeDate(PName pn) throws PExceptionNaming {
		return null;
	}

	public BigInteger encodeBigInteger(PName pn) throws PExceptionNaming {
		return null;
	}

	public BigDecimal encodeBigDecimal(PName pn) throws PExceptionNaming {
		return null;
	}

	public PName getNull() {
		return null;
	}

	public void setNullPName(Object o) throws PException {
	}

	public boolean supportDynamicComposite() {
		return false;
	}

	public boolean supportCompositeField(String fn, PType ft) {
		return false;
	}

	public boolean supportStaticComposite() {
		return false;
	}

	public PType getPType() {
		return null;
	}

	public void setPType(PType pt) {
	}

	// PRIVATE METHODS //
	//-----------------//

	private Class getClass(String classname, ClassLoader cl) throws PException {
		if (cl == null) {
			throw new PException("Impossible to load the class "
					+ classname + " without a classLoader(" + cl + ")");
		}
		try {
			return cl.loadClass(classname);
		} catch (Exception e) {
			throw new PException(e, "Impossible to load the class '"
					+ classname + "' with the class loader :" + cl);
		}
	}

	public synchronized PBinder findPBinder(String bcn,
											String clName,
											ClassLoader classLoader) throws PException {
		if (bcn == null || bcn.length() == 0)
			throw new JDOFatalInternalException(
					"Impossible to get the PNamingContext of the class " + clName
					+ ": Specified binder class is not valid: " + bcn);
		PBinder binder = (PBinder) binders.get(clName);
		if (binder != null)
			return binder;
		binder = nmf.getNamingManager(bcn, classLoader)
				.getPBinder(clName, bcn, classLoader,
				    mappingStructuresRule, binders, pnamingContexts);
		binder.setCacheManager(cache);
		binders.put(clName, binder);
		return binder;
	}

	protected synchronized PNameManager findPNameManager(String pnmcn,
														 String clName,
														 ClassLoader classLoader) throws PException {
		if (pnmcn == null || pnmcn.length() == 0)
			throw new JDOFatalInternalException(
					"Impossible to get the PNamingContext of the class " + clName
					+ ": Specified PNameManager class is not valid: " + pnmcn);
		PNameManager pnm = (PNameManager) pnamingContexts.get(clName);
		if (pnm != null)
			return pnm;
		NamingManager nm = nmf.getNamingManager(pnmcn, classLoader);
		if (nm.supportPNamingcontext()) {
			pnm = nm.getPNamingContext(clName, pnmcn, classLoader,
			    mappingStructuresRule, binders, pnamingContexts, mapper.getMetaInfoManager());
		} else {
			pnm = findPBinder(pnmcn, clName, classLoader);
		}
		pnamingContexts.put(clName, pnm);
		return pnm;
	}

	private class RefConfig implements PClassMapping.ReferenceConfigurator {

		Map classProperties = null;
		String mapperName = null;
		List gcms = null;
		ClassLoader classLoader = null;

		public RefConfig(Map classProperties, String mapperName, ClassLoader cl) {
			this.classProperties = classProperties;
			this.mapperName = mapperName;
			this.gcms = new ArrayList();
			this.classLoader = cl;
		}

		public List getGenClassMappings() {
			return gcms;
		}

		// IMPLEMENTATION OF THE PClassMapping.PNCGetter INTERFACE //
		//---------------------------------------------------------//


		public PNameCoder getPNameCoder(String sourceclassName,
										String refFieldName,
										String destclassName) {
			try {
				PNamingContext pnc = getPNamingContext(destclassName, classLoader);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG,
							"Assign the PNamingContext for the field " + sourceclassName
							+ "." + refFieldName + " which references the class "
							+ destclassName + ": " + pnc);
				logger.log(BasicLevel.DEBUG,
						"getPNameManager(" + destclassName + ") ==>" + pnc);
				return pnc;
			} catch (PException e) {
				throw new JDOFatalInternalException(
						"ERROR during the assignation of the PNameManager of the reference "
						+ refFieldName + " of the class " + sourceclassName, e);
			}
		}

		public PNameCoder getPNameCoder(String sourceclassName,
										String refFieldName,
										String[] genClassNames) {
			String path = JormPathHelper.getPath(
					sourceclassName, refFieldName, genClassNames);
			String hints = (String) classProperties.get(path);
			try {
				PNamingContext pnc = (PNamingContext) findPNameManager(hints, path, classLoader);
				logger.log(BasicLevel.DEBUG,
						"getPNameManager(" + path + ") ==>" + pnc);
				return pnc;
			} catch (PException e) {
				throw new JDOFatalInternalException(e.getMessage(), e);
			}
		}

		public PClassMapping getGenClassMapping(String sourceclassName,
												String refFieldName,
												String[] genClassNames) {

			String cn = JormPathHelper.getPath(
					sourceclassName, refFieldName, genClassNames);
			String pnmcn = (String) classProperties.get(cn);
			logger.log(BasicLevel.DEBUG, "getGenClassMapping(" + cn + ")");
			logger.log(BasicLevel.DEBUG, "\tpnmcn=" + pnmcn);
			try {
				PClassMapping gcm = (PClassMapping)
						Class.forName(getGCMClassName(mapperName)).newInstance();
				logger.log(BasicLevel.DEBUG, "\tgcm=" + gcm);
				PBinder pb = findPBinder(pnmcn, cn, classLoader);
				logger.log(BasicLevel.DEBUG, "\tpb=" + pb);
				gcm.setPBinder(pb);
				pb.setPClassMapping(gcm);
				gcms.add(gcm);
				return gcm;
			} catch (Exception e) {
				throw new JDOFatalInternalException(e.getMessage(), e);
			}
		}

		public PClassMapping getGenClassMapping(String sourceclassName,
												String refFieldName,
												String[] genClassNames,
												String destclassName) {
			String cn = JormPathHelper.getPath(
					sourceclassName, refFieldName, genClassNames);
			String pnmcn = (String) classProperties.get(cn);
			try {
				PClassMapping gcm = (PClassMapping)
						Class.forName(getGCMClassName(mapperName)).newInstance();
				logger.log(BasicLevel.DEBUG, "getGenClassMapping(" + cn + ")");
				logger.log(BasicLevel.DEBUG, "\tgcm=" + gcm);
				PBinder pb = findPBinder(pnmcn, cn, classLoader);
				logger.log(BasicLevel.DEBUG, "\tpb=" + pb);
				gcm.setPBinder(pb);
				pb.setPClassMapping(gcm);
				PNamingContext pnc = getPNamingContext(destclassName, classLoader);
				logger.log(BasicLevel.DEBUG, "\tpnc=" + pnc);
				((PClassMappingCtrl) gcm).setPNameCoder(pnc);
				gcms.add(gcm);
				return gcm;
			} catch (Exception e) {
				throw new JDOFatalInternalException(e.getMessage(), e);
			}
		}
	}

	private void loadJormMI(String className,
							Properties classProperties,
							ClassLoader cl) {
		JormManager m = (JormManager) mapper.getMetaInfoManager();
		synchronized (m) {
			if (m.getClass(className) != null) {
				return;
			}
			Set mos = null;
			String fn = classProperties.getProperty(Object2StringSerializer.JDO_FILE_NAME_PROP);
			try {
				mos = (Set) Object2StringSerializer.deserialize(fn, cl, logger);
				//mos = (Set) Object2JMIFileSerializer.deserialize(fn, cl, logger);
			} catch (JDOException e) {
				throw e;
			} catch (Exception e) {
				throw new JDOException("Impossible to load the jorm meta " +
						"information for the class '" + className + "' from the file '"
						+ fn + "' (You must use the same Speedo version for the"
						+ " enhancement and for the runtime): ", e);
			}

			for (Iterator it = mos.iterator(); it.hasNext();) {
				Object o = it.next();
				if (o instanceof BasicClass) {
					BasicClass c = (BasicClass) o;
					Package s = m.createPackage(((Package) c.getParent()).getName());
					s.addClass(c);
					c.setParent(s);
					c.setLogger(m.getLogger());
					logger.log(BasicLevel.DEBUG,
							"Jorm Meta Object Class " + c.getFQName() + " loaded");
				} else if (o instanceof BasicCompositeName) {
					BasicCompositeName cn = (BasicCompositeName) o;
					Package s = m.createPackage(((Package) cn.getParent()).getName());
					s.addCompositeName(cn);
					cn.setLogger(m.getLogger());
					logger.log(BasicLevel.DEBUG,
							"Jorm Meta Object CompositeName " + cn.getFQName() + " loaded");
				} else
					throw new JDOException("Umanaged Jorm Meta Object " + o);
			}
			if (m.getClass(className) == null) {
				throw new JDOException(
						"Internal ERROR: No meta information found about the persistent class '"
						+ className + "' in the file '" + fn + "'.");
			}
		}
	}

	private String getGCMClassName(String mapperName) {
		if (mapperName.startsWith("rdb")) {
			return "org.objectweb.jorm.mapper.rdb.genclass.RdbGenClassMapping";
		} else {
			throw new JDOFatalInternalException("Umanaged mapper: " + mapperName);
		}
	}

	private ClassLoader getClassLoader(Class c) {
		ClassLoader cl = c.getClassLoader();
		if (cl == null) {
			cl = this.getClass().getClassLoader();
			if (cl == null) {
				cl = ClassLoader.getSystemClassLoader();
				logger.log(BasicLevel.DEBUG, "Use the system class loader for the class '"
							+ c.getName() + "': " + cl);
			} else {
				logger.log(BasicLevel.DEBUG, "Use the Speedo class loader for the class '"
					+ c.getName() + "': " + cl);
			}
		} else {
			logger.log(BasicLevel.DEBUG, "Use the Application class loader for the class '"
				+ c.getName() + "': " + cl);
		}
		return cl;
	}
}
