package org.objectweb.speedo.runtime.query.parser;

import junit.framework.TestCase;

import java.io.CharArrayReader;

import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.speedo.query.parser.SpeedoQL;

/**
 * Speedo QL Test
 * Test a list of valid filters.
 */
public class SpeedoQLTest extends TestCase {
    final static String filters[] = {
        "toto.titi",
        "toto.titi()",
        "this.toto",
        "this.titi()",
        "(+78)",
        "(-89458.457)",
        "(~788)",
        "((~abc)|| tyty)",
        "(!rere && toto+7)",
        "((boolean)john)",
        "(a < b > c <= d >= e | f || g & h && i == j != k + l - m / n * o ^ p.jotp())",
        "((byte)b && (short)s || (int)i <= (long)l == (char)c * (float)f / (double)d == (true) && (false | null))",
        "null",
        "((char)(byte)(short)b || (char)(java.lang.String)s)",
        "my.toto((q<b) + 9)",
        "(house.b())",
        "((((g)a < b.g())))",
        "(+4 < t)",
        "((-4+9*8) == toto.titi.titi.ktkt(one))",
        "((a < b.house()) || (c.name > ((d*4)/100)))",
        "toto(tutu).titi(tata)",
        "toto(titi)",
        "((1 < 5 > 3) || (4  && 5 <= (934 == 4)))",
        "(((1 < 5) > 3) || ((4  && 5) <= (934 == 4)))",
        "(true == a)",
        "true == a",
        "(a+b).method()",
        "(salary > 3000)",
        "(salary > sal)",
        "(dept.name == dep)",
        "(emps.contains(emp) & emp.salary > sal)",
        "(depts.contains(name))",
        "((getEmpId () >= 1) && (getEmpId () <= 10))",
        "a == \"Paul\"",
        "a >= 1 && b <= 10",
        "x == param",
        "x == 4",
        "x == param1 && y == param2",
        "reviewers.contains(person) && person == Employee",
        "reviewers.contains(person) && person == person"
		//"toto(titi, tatat)",
    };

    private final static String LOGGER_NAME
            = "org.objectweb.speedo.rt.query.parser.speedoql";

    private static Logger logger = null;

    static {
        logger = Monolog.initialize().getLogger(LOGGER_NAME);
    }


    public SpeedoQLTest(String name) {
        super(name);
    }

    public void testAll() {
        logger.log(BasicLevel.DEBUG, "SpeedoQL Test - javacc grammar");
        for (int i = 0; i < filters.length; i++) {
            String filter = filters[i];
            logger.log(BasicLevel.DEBUG, "Testing: " + filter);
            try {
                filter = '(' + filter + ')';
                new SpeedoQL(new CharArrayReader(filter.toCharArray())).SpeedoQL();
            } catch (Throwable e) {
                logger.log(BasicLevel.ERROR, "Parsing of: " + filter, e);
                fail("Parsing of: " + filter);
            }

        }
    }
}
